@@ -1,3 +1,8 @@
+(* Generated code should depend on the environment in scope as little as
+   possible.  E.g. rather than [foo = []] do [match foo with [] ->], to eliminate the
+   use of [=].  It is especially important to not use polymorphic comparisons, since we
+   are moving more and more to code that doesn't have them in scope. *)
+
 module List = ListLabels
 open Printf
 open Camlp4.PreCast
@@ -37,8 +42,40 @@ module Inspect = struct
   let fields ty = List.map (Ast.list_of_ctyp ty []) ~f:field
 end
 
+let generate_at_least_once rec_ ~f ~combine typedefs =
+  if not rec_ then
+    failwith "nonrec is not compatible with the `fields' preprocessor";
+  let rec aux = function
+    | Ast.TyDcl (_loc, ty_name, tps, rhs, _) -> f _loc ~ty_name ~tps ~rhs
+    | Ast.TyAnd (_loc, td1, td2) -> (
+      match aux td1, aux td2 with
+      | `Ok str1, `Ok str2 -> `Ok (combine _loc str1 str2)
+      | `Ok str1, `Error _ -> `Ok str1
+      | `Error _, `Ok str2 -> `Ok str2
+      | `Error _, `Error _ ->
+        `Error "'with fields' can only be applied on type definitions in which at \
+                least one type definition is a record"
+    )
+    | Ast.TyNil _loc ->
+      `Error "'with fields': unexpected TyNil without a TyAnd somewhere around!!"
+    | _ -> assert false in
+
+  match aux typedefs with
+  | `Ok res -> res
+  | `Error s -> failwith s
+
 let raise_unsupported () =
-  failwith "Unsupported use of fields (you can only use it on records)."
+  `Error "Unsupported use of fields (you can only use it on records)."
+
+let perm _loc private_ =
+  match private_ with
+  | true -> <:ctyp< [< `Read ] >>
+  | false -> <:ctyp< [< `Read | `Set_and_create ] >>
+
+let field_t _loc private_ =
+  match private_ with
+  | false -> <:ctyp< Fieldslib.Field.t >>
+  | true -> <:ctyp< Fieldslib.Field.readonly_t >>
 
 module Gen_sig = struct
   let apply_type _loc ~ty_name ~tps =
@@ -48,15 +85,15 @@ module Gen_sig = struct
 
   let label_arg _loc name ty = Ast.TyLab (_loc, name, ty)
 
-  let field_arg _loc ~record f = fun (name, _m, ty) ->
+  let field_arg _loc ~private_ ~record f = fun (name, _m, ty) ->
     label_arg _loc name (
-      f ~field: <:ctyp< Fieldslib.Field.t $record$ $ty$ >> ~ty)
+      f ~field: <:ctyp< $field_t _loc private_$ $record$ $ty$ >> ~ty)
   ;;
 
   let create_fun ~ty_name ~tps _loc ty =
     let record = apply_type _loc ~ty_name ~tps in
     let fields = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty ->
+    let f = field_arg _loc ~private_:false ~record (fun ~field ~ty ->
       let create_f = <:ctyp< 'input__ -> ( $ty$ ) >> in
       <:ctyp< $field$ -> 'compile_acc__ -> ($create_f$ * 'compile_acc__) >>
     )  in
@@ -69,10 +106,10 @@ module Gen_sig = struct
   ;;
 
 
-  let fold_fun ~ty_name ~tps _loc ty =
+  let fold_fun ~private_ ~ty_name ~tps _loc ty =
     let record  = apply_type _loc ~ty_name ~tps in
     let fields  = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:_ ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:_ ->
       <:ctyp< 'acc__ -> $field$ -> 'acc__ >>) in
     let types   = List.map fields ~f in
     let init_ty = label_arg _loc "init" <:ctyp< 'acc__ >> in
@@ -93,40 +130,40 @@ module Gen_sig = struct
 
 
 
-  let bool_fun fun_name ~ty_name ~tps _loc ty =
+  let bool_fun fun_name ~private_ ~ty_name ~tps _loc ty =
     let record    = apply_type _loc ~ty_name ~tps in
     let fields    = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:_ ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:_ ->
       <:ctyp< $field$ -> bool >> ) in
     let types   = List.map fields ~f in
     let t       = Create.lambda_sig _loc types <:ctyp< bool >> in
     <:sig_item< value $lid:fun_name$ : $t$ >>
   ;;
 
-  let iter_fun ~ty_name ~tps _loc ty =
+  let iter_fun ~private_ ~ty_name ~tps _loc ty =
     let record    = apply_type _loc ~ty_name ~tps in
     let fields    = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:_ ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:_ ->
       <:ctyp< $field$ -> unit >>) in
     let types  = List.map fields ~f in
     let t     = Create.lambda_sig _loc types <:ctyp< unit >> in
     <:sig_item< value iter : $t$ >>
   ;;
 
-  let direct_iter_fun ~ty_name ~tps _loc ty =
+  let direct_iter_fun ~private_ ~ty_name ~tps _loc ty =
     let record    = apply_type _loc ~ty_name ~tps in
     let fields    = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:field_ty ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:field_ty ->
       <:ctyp< $field$ -> $record$ -> $field_ty$ -> unit >>) in
     let types  = List.map fields ~f in
     let t     = Create.lambda_sig _loc (record :: types) <:ctyp< unit >> in
     <:sig_item< value iter : $t$ >>
   ;;
 
-  let direct_fold_fun ~ty_name ~tps _loc ty =
+  let direct_fold_fun ~private_ ~ty_name ~tps _loc ty =
     let record  = apply_type _loc ~ty_name ~tps in
     let fields  = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:field_ty ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:field_ty ->
       <:ctyp< 'acc__ -> $field$ -> $record$ -> $field_ty$ -> 'acc__ >>) in
     let types   = List.map fields ~f in
     let init_ty = label_arg _loc "init" <:ctyp< 'acc__ >> in
@@ -135,10 +172,10 @@ module Gen_sig = struct
     <:sig_item< value fold : $t$ >>
   ;;
 
-  let to_list_fun ~ty_name ~tps _loc ty =
+  let to_list_fun ~private_ ~ty_name ~tps _loc ty =
     let record    = apply_type _loc ~ty_name ~tps in
     let fields    = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty:_ ->
+    let f = field_arg _loc ~private_ ~record (fun ~field ~ty:_ ->
         <:ctyp< $field$ -> 'elem__ >>)
     in
     let types = List.map fields ~f in
@@ -149,7 +186,7 @@ module Gen_sig = struct
   let map_fun ~ty_name ~tps _loc ty =
     let record    = apply_type _loc ~ty_name ~tps in
     let fields    = Inspect.fields ty in
-    let f = field_arg _loc ~record (fun ~field ~ty ->
+    let f = field_arg _loc ~private_:false ~record (fun ~field ~ty ->
       <:ctyp< $field$ -> $ty$ >>) in
     let types = List.map fields ~f in
     let t     = Create.lambda_sig _loc (types) record in
@@ -157,7 +194,7 @@ module Gen_sig = struct
   ;;
 
 
- let map_poly ~ty_name ~tps _loc _ =
+ let map_poly ~private_ ~ty_name ~tps _loc _ =
    let record    = apply_type _loc ~ty_name ~tps in
    let tps_names =
      List.map
@@ -176,28 +213,30 @@ module Gen_sig = struct
       in
       <:ctyp<'$lid:loop 0$>>
     in
+    let perm = perm _loc private_ in
     let t =
-      <:ctyp< Fieldslib.Field.user $record$ $fresh_variable$ -> list $fresh_variable$  >>
+      <:ctyp< Fieldslib.Field.user $perm$ $record$ $fresh_variable$ -> list $fresh_variable$  >>
     in
     <:sig_item< value map_poly : $t$ >>
  ;;
 
 
 
-  let record ~ty_name ~tps _loc ty =
+  let record ~private_ ~ty_name ~tps _loc ty =
     let fields = Inspect.fields ty in
     let record_ty = apply_type _loc ~ty_name ~tps in
     let conv_field (res_getset, res_fields) (name, m, ty) =
       let getter = <:sig_item< value $lid:name$ : $record_ty$ -> $ty$ >> in
       let field  =
-        <:sig_item< value $lid:name$ : Fieldslib.Field.t $record_ty$ $ty$ >>
+        <:sig_item< value $lid:name$ : $field_t _loc private_$ $record_ty$ $ty$ >>
       in
-      match m with
-      | `Immutable ->
+      match m, private_ with
+      | `Immutable, _
+      | `Mutable, true ->
           ( <:sig_item< $getter$ ; $res_getset$ >> ,
             <:sig_item< $field$ ; $res_fields$ >>
           )
-      | `Mutable ->
+      | `Mutable, false ->
           let setter    =
             <:sig_item< value $lid:"set_" ^ name$ : $record_ty$ -> $ty$ -> unit >> in
           ( <:sig_item< $getter$ ; $setter$ ;  $res_getset$ >> ,
@@ -209,57 +248,61 @@ module Gen_sig = struct
     in
     let create_fun = create_fun ~ty_name ~tps _loc ty in
     let simple_create_fun = simple_create_fun ~ty_name ~tps _loc ty in
-    if ty_name = "t" then
-      let iter        = iter_fun ~ty_name ~tps _loc ty in
-      let fold        = fold_fun ~ty_name ~tps _loc ty in
-      let map         = map_fun ~ty_name ~tps _loc ty in
-      let map_poly         = map_poly ~ty_name ~tps _loc ty in
-      let and_f       = bool_fun "for_all" ~ty_name ~tps _loc ty in
-      let or_f        = bool_fun "exists" ~ty_name ~tps _loc ty in
-      let to_list     = to_list_fun ~ty_name ~tps _loc ty in
-      let direct_iter = direct_iter_fun ~ty_name ~tps _loc ty in
-      let direct_fold = direct_fold_fun ~ty_name ~tps _loc ty in
-      <:sig_item< $getters_and_setters$ ;
-          module Fields : sig
-            value names : list string ;
-            $fields$ ;
-            $fold$ ;
-            $create_fun$ ; $simple_create_fun$ ; $iter$ ; $map$ ; $map_poly$ ; $and_f$ ; $or_f$ ; $to_list$ ;
-            module Direct : sig
-              $direct_iter$ ;
-              $direct_fold$ ;
-            end ;
-          end
-      >>
-    else
-      <:sig_item< $getters_and_setters$ ;
-          module Fields : sig
-            $fields$
-          end
-      >>
+    let fields_module = if ty_name = "t" then "Fields" else "Fields_of_" ^ ty_name in
+    let iter        = iter_fun ~private_ ~ty_name ~tps _loc ty in
+    let fold        = fold_fun ~private_ ~ty_name ~tps _loc ty in
+    let map         = map_fun ~ty_name ~tps _loc ty in
+    let map_poly    = map_poly ~private_ ~ty_name ~tps _loc ty in
+    let and_f       = bool_fun "for_all" ~private_ ~ty_name ~tps _loc ty in
+    let or_f        = bool_fun "exists" ~private_ ~ty_name ~tps _loc ty in
+    let to_list     = to_list_fun ~private_ ~ty_name ~tps _loc ty in
+    let direct_iter = direct_iter_fun ~private_ ~ty_name ~tps _loc ty in
+    let direct_fold = direct_fold_fun ~private_ ~ty_name ~tps _loc ty in
+    <:sig_item< $getters_and_setters$ ;
+        module $uid:fields_module$ : sig
+          value names : list string ;
+          $fields$ ;
+          $fold$ ;
+          $ if private_
+            (* The ['perm] phantom type prohibits first-class fields from mutating or
+               creating private records, so we can expose them (and fold, etc.).
+
+               However, we still can't expose functions that explicitly create private
+               records.
+            *)
+            then <:sig_item< >>
+            else <:sig_item< $create_fun$ ; $simple_create_fun$ ; $map$ ; >>
+          $ ;
+          $iter$ ; $and_f$ ; $or_f$ ; $to_list$ ; $map_poly$ ;
+          module Direct : sig
+            $direct_iter$ ;
+            $direct_fold$ ;
+          end ;
+        end
+    >>
   ;;
 
   let fields_of_ty_sig _loc ~ty_name ~tps ~rhs =
-    let unsupported = (fun _ _ -> raise_unsupported ()) in
-    Gen.switch_tp_def
-      ~alias:unsupported
-      ~sum:unsupported
-      ~variants:unsupported
-      ~mani:unsupported
-      ~nil:(fun _ -> raise_unsupported ())
-      ~record:(record ~ty_name ~tps)
-      rhs
+    match rhs with
+    | <:ctyp@loc< $_$ == private { $flds$ } >>
+    | <:ctyp@loc< private { $flds$ } >> ->
+      `Ok (record ~ty_name ~private_:true ~tps loc flds)
+    | <:ctyp@loc< $_$ == { $flds$ } >>
+    | <:ctyp@loc< { $flds$ } >> ->
+      `Ok (record ~ty_name ~private_:false ~tps loc flds)
+    | _ -> raise_unsupported ()
+
+  let generate rec_ typedefs =
+    generate_at_least_once
+      rec_
+      ~f:fields_of_ty_sig
+      ~combine:(fun _loc item1 item2 -> <:sig_item< $item1$; $item2$; >>)
+      typedefs
 
-  let generate = function
-    | Ast.TyDcl (_loc, ty_name, tps, rhs, _) -> fields_of_ty_sig _loc ~ty_name ~tps ~rhs
-    | Ast.TyAnd (_loc, _, _) as tds    ->
-        ignore (_loc, tds);
-        failwith "Not supported"
-    | _                             -> assert false
 end
 
 module Gen_struct = struct
-  let fields _loc ty =
+  let fields ~private_ _loc ty =
     let fields = Inspect.fields ty in
     let rec_id =
       match fields with
@@ -269,8 +312,11 @@ module Gen_struct = struct
     let conv_field (res_getset, res_fields) (name, m, field_ty) =
       let getter = <:str_item< value $lid:name$ _r__ = _r__.$lid:name$ >> in
       let setter, setter_field =
-        match m with
-        | `Mutable ->
+        match m, private_ with
+        | `Mutable, true ->
+          <:str_item< >>,
+          <:expr< Some (fun _ _ -> failwith "invalid call to a setter of a private type") >>
+        | `Mutable, false ->
             let setter =
               <:str_item<
                 value $lid:"set_" ^ name$ _r__ v__ = _r__.$lid:name$ := v__
@@ -278,7 +324,7 @@ module Gen_struct = struct
             in
             let setter_field = <:expr< Some $lid:"set_" ^ name$ >> in
             setter, setter_field
-        | `Immutable -> <:str_item< >>, <:expr< None >>
+        | `Immutable, _ -> <:str_item< >>, <:expr< None >>
       in
       let field  =
         let e =
@@ -287,15 +333,22 @@ module Gen_struct = struct
               Ast.ExId (_loc, Ast.IdLid (_loc, "v__"))),
             rec_id)
         in
-        let fset = <:expr< fun _r__ v__ -> $e$ >> in
+        let fset =
+          match private_ with
+          | true ->
+            <:expr< fun _ _ -> failwith "Invalid call to an fsetter of a private type" >>
+          | false -> <:expr< fun _r__ v__ -> $e$ >>
+        in
+        let perm = perm _loc private_ in
         <:str_item<
-          value $lid:name$ =
-            ( { Fieldslib.Field.
-              name    = $str:name$;
-              getter  = $lid:name$;
-              setter  = $setter_field$;
-              fset    = $fset$;
-            } : Fieldslib.Field.t _ $field_ty$ )
+          value $lid:name$ : Fieldslib.Field.t_with_perm $perm$ _ $field_ty$ =
+              Fieldslib.Field.Field { Fieldslib.Field.For_generated_code.
+                force_variance = (fun (_ : $perm$) -> ());
+                name = $str:name$;
+                getter = $lid:name$;
+                setter = $setter_field$;
+                fset = $fset$;
+              }
         >>
       in
       ( <:str_item< $getter$ ; $setter$ ; $res_getset$ >>,
@@ -351,7 +404,6 @@ module Gen_struct = struct
     let f    = Create.lambda _loc (patterns @ [ <:patt< compile_acc__ >> ]) body in
     <:str_item<
       value make_creator = $f$;
-      value _ = make_creator
     >>
   ;;
 
@@ -366,7 +418,6 @@ module Gen_struct = struct
     let f    = Create.lambda _loc patterns f  in
     <:str_item<
       value create = $f$;
-      value _ = create
     >>
   ;;
 
@@ -382,7 +433,6 @@ module Gen_struct = struct
       ( init :: patterns ) body in
     <:str_item<
       value fold = $lambda$;
-      value _ = fold
     >>
   ;;
 
@@ -396,7 +446,6 @@ module Gen_struct = struct
     let lambda = Create.lambda _loc patterns body in
     <:str_item<
       value for_all = $lambda$;
-      value _ = for_all
     >>
   ;;
 
@@ -410,7 +459,6 @@ module Gen_struct = struct
     let lambda = Create.lambda _loc patterns body in
     <:str_item<
       value exists = $lambda$;
-      value _ = exists
     >>
   ;;
 
@@ -427,7 +475,6 @@ module Gen_struct = struct
         (patterns) body in
     <:str_item<
       value iter = $lambda$;
-      value _ = iter
     >>
   ;;
 
@@ -446,7 +493,6 @@ module Gen_struct = struct
     let lambda     = Create.lambda _loc ( <:patt< record__ >> :: patterns) body in
     <:str_item<
       value iter = $lambda$;
-      value _ = iter
     >>
   ;;
 
@@ -465,7 +511,6 @@ module Gen_struct = struct
       ( <:patt< record__ >> :: init :: patterns ) body in
     <:str_item<
       value fold = $lambda$;
-      value _ = fold
     >>
   ;;
 
@@ -478,7 +523,6 @@ module Gen_struct = struct
     let f    = Create.lambda _loc patterns body in
     <:str_item<
       value map = $f$;
-      value _ = map
     >>
   ;;
 
@@ -492,7 +536,6 @@ module Gen_struct = struct
     let f        = Create.lambda _loc patterns body in
     <:str_item<
       value to_list = $f$;
-      value _ = to_list
     >>
   ;;
 
@@ -509,73 +552,62 @@ module Gen_struct = struct
     in
     <:str_item<
       value map_poly record__ = $body$;
-      value _ = map_poly
     >>
   ;;
 
-  let record ~record_name _loc ty =
-    let getter_and_setters, fields = fields _loc ty in
+  let record ~private_ ~record_name _loc ty =
+    let getter_and_setters, fields = fields ~private_ _loc ty in
     let create   = creation_fun _loc record_name ty in
     let simple_create   = simple_creation_fun _loc record_name ty in
     let names =
       List.fold_right (Inspect.field_names ty) ~init:<:expr< [ ] >> ~f:(fun head tail ->
         <:expr< [ $str:head$ :: $tail$ ] >>)
     in
-    if record_name = "t" then
-      let iter        = iter_fun ~record_name _loc ty in
-      let fold        = fold_fun ~record_name _loc ty in
-      let map         = map_fun ~record_name _loc ty in
-      let map_poly    = map_poly ~record_name _loc ty in
-      let andf        = and_fun ~record_name _loc ty in
-      let orf         = or_fun ~record_name _loc ty in
-      let direct_iter = direct_iter_fun ~record_name _loc ty in
-      let direct_fold = direct_fold_fun ~record_name _loc ty in
-      let to_list     = to_list_fun ~record_name _loc ty in
-      <:str_item<
-        $getter_and_setters$ ;
-        module Fields = struct
-          value names = $names$ ;
-          value _ = names ;
-          $fields$ ;
-          $create$ ; $simple_create$ ; $iter$ ; $fold$ ; $map$ ; $map_poly$ ; $andf$ ; $orf$ ; $to_list$ ;
-          module Direct = struct
-            $direct_iter$ ;
-            $direct_fold$ ;
-          end ;
-        end
-      >>
-    else
-      <:str_item<
-        $getter_and_setters$ ;
-        module Fields = struct
-          $fields$ ;
-        end
-      >>
+    let fields_module = if record_name = "t" then "Fields" else "Fields_of_" ^ record_name in
+    let iter        = iter_fun ~record_name _loc ty in
+    let fold        = fold_fun ~record_name _loc ty in
+    let map         = map_fun ~record_name _loc ty in
+    let map_poly    = map_poly ~record_name _loc ty in
+    let andf        = and_fun ~record_name _loc ty in
+    let orf         = or_fun ~record_name _loc ty in
+    let direct_iter = direct_iter_fun ~record_name _loc ty in
+    let direct_fold = direct_fold_fun ~record_name _loc ty in
+    let to_list     = to_list_fun ~record_name _loc ty in
+    <:str_item<
+      $getter_and_setters$ ;
+      module $uid:fields_module$ = struct
+        value names = $names$ ;
+        $fields$;
+        $ if private_
+          then <:str_item< >>
+          else <:str_item< $create$ ; $simple_create$; $map$; >>
+        $ ;
+        $iter$ ; $fold$ ; $map_poly$ ;
+        $andf$ ; $orf$ ; $to_list$ ;
+        module Direct = struct
+          $direct_iter$ ;
+          $direct_fold$ ;
+        end ;
+      end
+    >>
   ;;
 
-  let mani ~record_name ty =
-    match ty with
-    | <:ctyp@loc< { $x$ } >> ->
-      record ~record_name loc x
-    | _ -> failwith "the right hand side of the manifest must be a record"
+  let fields_of_ty _loc ~ty_name:record_name ~tps:_ ~rhs =
+    match rhs with
+    | <:ctyp@loc< $_$ == private { $flds$ } >>
+    | <:ctyp@loc< private { $flds$ } >> ->
+      `Ok (record ~record_name ~private_:true loc flds)
+    | <:ctyp@loc< $_$ == { $flds$ } >>
+    | <:ctyp@loc< { $flds$ } >> ->
+      `Ok (record ~record_name ~private_:false loc flds)
+    | _ -> raise_unsupported ()
 
-  let fields_of_ty _loc ~record_name ~tps:_ ~rhs =
-    let unsupported = (fun _ _ -> raise_unsupported ()) in
-    Gen.switch_tp_def
-      ~alias:    unsupported
-      ~sum:      unsupported
-      ~variants: unsupported
-      ~mani:     (fun (_:Loc.t) _tp1 tp2 -> mani ~record_name tp2)
-      ~nil:      (fun _ -> raise_unsupported ())
-      ~record:   (record ~record_name)
-      rhs
-
-  let generate = function
-    | Ast.TyDcl (_loc, name, tps, rhs, _) -> fields_of_ty _loc ~record_name:name ~tps ~rhs
-    | Ast.TyAnd (_loc, _, _) as tds ->
-        ignore (_loc, tds);
-        failwith "Not supported"
-    | _                             -> assert false
+  let generate rec_ typedefs =
+    generate_at_least_once
+      rec_
+      ~f:fields_of_ty
+      ~combine:(fun _loc item1 item2 -> <:str_item< $item1$; $item2$; >>)
+      typedefs
 end
 
 let () = add_generator "fields" Gen_struct.generate
